from typing import Optional, Dict, Any, Union

from opentelemetry.context import Context
from opentelemetry.metrics import Histogram


class TelemetryHistogram:
    """
    This class wraps the OpenTelemetry Histogram class and adds the supplied attributes every time a metric is recorded

    """

    def __init__(
        self, *, histogram: Histogram, attributes: Optional[Dict[str, Any]]
    ) -> None:
        assert histogram
        self._histogram: Histogram = histogram
        self._attributes: Optional[Dict[str, Any]] = attributes

    def record(
        self,
        amount: Union[int, float],
        attributes: Optional[Dict[str, Any]] = None,
        context: Optional[Context] = None,
    ) -> None:
        attributes = attributes or {}
        attributes.update(self._attributes or {})

        self._histogram.record(amount=amount, attributes=attributes, context=context)



import os
from shutil import rmtree

from pathlib import Path
from typing import Protocol, runtime_checkable


from helixcore.utilities.dynamic_class_loader.v1.dynamic_class_loader import (
    DynamicClassLoader,
)


# Create a base protocol/class for testing
@runtime_checkable
class Animal(Protocol):
    def make_sound(self) -> str:
        pass


def test_dynamic_class_loader() -> None:
    """Test dynamic class loading functionality."""

    # Create test module files
    dog_content = """
from typing import Protocol, runtime_checkable

@runtime_checkable
class Animal(Protocol):
    def make_sound(self) -> str:
        ...

class Dog:
    def make_sound(self) -> str:
        return "Woof!"
    """

    cat_content = """
from typing import Protocol, runtime_checkable

@runtime_checkable
class Animal(Protocol):
    def make_sound(self) -> str:
        ...

class Cat:
    def make_sound(self) -> str:
        return "Meow!"
    """

    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("../temp")
    if os.path.isdir(temp_folder):
        rmtree(temp_folder)
    os.makedirs(temp_folder)

    # Write files
    with open(os.path.join(temp_folder, "dog.py"), "w") as f:
        f.write(dog_content)

    with open(os.path.join(temp_folder, "cat.py"), "w") as f:
        f.write(cat_content)

    # Create a DynamicClassLoader instance
    loader = DynamicClassLoader(base_class=Animal, folder_path=temp_folder)  # type: ignore[type-abstract]

    # Find subclasses
    subclasses = loader.find_subclasses()

    # Verify subclasses
    assert len(subclasses) == 2

    # Verify class names
    class_names = {cls.__name__ for cls in subclasses}
    assert class_names == {"Dog", "Cat"}

    # Verify method functionality
    for cls in subclasses:
        instance = cls()
        assert hasattr(instance, "make_sound")
        assert callable(instance.make_sound)


def test_dynamic_class_loader_empty_dir() -> None:
    """Test behavior with an empty directory."""
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("../temp")
    if os.path.isdir(temp_folder):
        rmtree(temp_folder)
    os.makedirs(temp_folder)

    loader = DynamicClassLoader(base_class=Animal, folder_path=temp_folder)  # type: ignore[type-abstract]
    subclasses = loader.find_subclasses()

    assert len(subclasses) == 0


def test_dynamic_class_loader_init() -> None:
    """Test the initialization of DynamicClassLoader."""
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("../temp")
    if os.path.isdir(temp_folder):
        rmtree(temp_folder)
    os.makedirs(temp_folder)

    loader = DynamicClassLoader(base_class=Animal, folder_path=temp_folder)  # type: ignore[type-abstract]

    assert loader.base_class == Animal
    assert loader.folder_path == Path(temp_folder)



from typing import Optional, Dict, Any, Union

from opentelemetry.context import Context
from opentelemetry.metrics import UpDownCounter


class TelemetryUpDownCounter:
    """
    This class wraps the OpenTelemetry UpDownCounter class and adds the supplied attributes every time a metric is recorded

    """

    def __init__(
        self, *, counter: UpDownCounter, attributes: Optional[Dict[str, Any]]
    ) -> None:
        assert counter
        self._counter: UpDownCounter = counter
        self._attributes: Optional[Dict[str, Any]] = attributes

    def add(
        self,
        amount: Union[int, float],
        attributes: Optional[Dict[str, Any]] = None,
        context: Optional[Context] = None,
    ) -> None:
        attributes = attributes or {}
        attributes.update(self._attributes or {})

        self._counter.add(amount=amount, attributes=attributes, context=context)



import os
import inspect
import importlib.util
from pathlib import Path
from typing import Type, List, Any
import abc


class DynamicClassLoader[T]:
    def __init__(self, base_class: Type[T], folder_path: Path | str):
        self.base_class: Type[T] = base_class
        self.folder_path: Path | str = folder_path

    def find_subclasses(self) -> List[Type[T]]:
        subclasses = []

        # Iterate through files in the specified folder
        for filename in os.listdir(self.folder_path):
            if filename.endswith(".py") and filename != "__init__.py":
                module_name = filename[:-3]
                file_path = os.path.join(self.folder_path, filename)

                # Dynamically import the module
                spec = importlib.util.spec_from_file_location(module_name, file_path)
                assert spec
                module = importlib.util.module_from_spec(spec)
                assert module
                assert spec.loader
                spec.loader.exec_module(module)

                # Inspect classes in the module
                for name, obj in inspect.getmembers(module, inspect.isclass):
                    # Skip if:
                    # 1. Not a subclass of base_class
                    # 2. Is the base class itself
                    # 3. Is a Protocol
                    # 4. Is an Abstract Base Class
                    # 5. Is a runtime checkable Protocol
                    is_skippable = (
                        not issubclass(obj, self.base_class)
                        or obj is self.base_class
                        or _is_protocol(obj)
                        or _is_abstract_base_class(obj)
                    )

                    if not is_skippable:
                        subclasses.append(obj)

        return subclasses


def _is_protocol(cls: Type[Any]) -> bool:
    """
    Check if a class is a Protocol

    Handles different ways Protocols can be defined:
    1. Runtime checkable Protocols
    2. Structural Protocols
    """
    return (
        # Check for runtime_checkable Protocols
        hasattr(cls, "_is_protocol")
        or hasattr(cls, "__protocol__")
        or
        # Check for abc.ABC with @runtime_checkable
        (isinstance(cls, type) and hasattr(cls, "__protocol__"))
    )


def _is_abstract_base_class(cls: Type[Any]) -> bool:
    """
    Check if a class is an Abstract Base Class
    """
    return (
        # Check if it's an ABC
        issubclass(cls, abc.ABC)
        or
        # Check for abstract methods
        any(
            getattr(getattr(cls, method), "__isabstractmethod__", False)
            for method in dir(cls)
        )
    )



# noinspection Mypy
from typing import Any

from setuptools import setup, find_packages
from os import path, getcwd

# from https://packaging.python.org/tutorials/packaging-projects/

# noinspection SpellCheckingInspection
package_name = "helixcore"

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    with open(path.join(getcwd(), "VERSION")) as version_file:
        version = version_file.read().strip()
except IOError:
    raise


def fix_setuptools() -> None:
    """Work around bugs in setuptools.

    Some versions of setuptools are broken and raise SandboxViolation for normal
    operations in a virtualenv. We therefore disable the sandbox to avoid these
    issues.
    """
    try:
        from setuptools.sandbox import DirectorySandbox

        # noinspection PyUnusedLocal
        def violation(operation: Any, *args: Any, **_: Any) -> None:
            print("SandboxViolation: %s" % (args,))

        DirectorySandbox._violation = violation
    except ImportError:
        pass


# Fix bugs in setuptools.
fix_setuptools()


# classifiers list is here: https://pypi.org/classifiers/

# create the package setup
setup(
    name=package_name,
    version=version,
    author="Imran",
    author_email="imran.qureshi@icanbwell.com",
    description="helixcore",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/icanbwell/helix-core",
    packages=find_packages(),
    install_requires=[
        "fhir.resources>=7.1.0,<8",
        "dataclasses-json>=0.6.7",
        "boto3>=1.34.140",
        "helix.fhir.client.sdk>=3.0.27",
        "furl>=2.1.3",
        "sqlparse>=0.5.3",
        "pymysql>=1.1.1",
        "aiomysql>=0.2.0",
        "aiohttp>=3.11.13",
        "structlog>=23.1.0",
        "opentelemetry-api>=1.30.0",
        "opentelemetry-sdk>=1.30.0",
        "opentelemetry-exporter-otlp>=1.30.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    dependency_links=[],
    include_package_data=True,
    zip_safe=False,
    package_data={"helixcore": ["py.typed"]},
)



from os import environ
from types import TracebackType
from typing import Optional, List, Dict, Any

from helixcore.event_loggers.event_logger import EventLogger
from helixcore.logger.log_level import LogLevel
from helixcore.logger.yarn_logger import get_logger


class MlFlowConfig:
    def __init__(
        self,
        mlflow_tracking_url: str,
        artifact_url: str,
        experiment_name: str,
        flow_run_name: str,
        parameters: Dict[str, Any],
    ):
        self.mlflow_tracking_url = mlflow_tracking_url
        self.artifact_url = artifact_url
        self.experiment_name = experiment_name
        self.flow_run_name = flow_run_name
        self.parameters = parameters

    def clone(self) -> "MlFlowConfig":
        return MlFlowConfig(
            mlflow_tracking_url=self.mlflow_tracking_url,
            artifact_url=self.artifact_url,
            experiment_name=self.experiment_name,
            flow_run_name=self.flow_run_name,
            parameters=self.parameters.copy(),
        )


class ProgressLogger:
    def __init__(
        self,
        event_loggers: Optional[List[EventLogger]] = None,
        mlflow_config: Optional[MlFlowConfig] = None,
    ) -> None:
        self.logger = get_logger(__name__)
        self.event_loggers: Optional[List[EventLogger]] = event_loggers
        self.mlflow_config: Optional[MlFlowConfig] = mlflow_config
        system_log_level_text: Optional[str] = environ.get("LOGLEVEL")
        self.system_log_level: Optional[LogLevel] = (
            LogLevel.from_str(system_log_level_text)
            if system_log_level_text is not None
            else None
        )

    def __enter__(self) -> "ProgressLogger":
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        pass

    def start_mlflow_run(self, run_name: str, is_nested: bool = True) -> None:
        pass

    # noinspection PyMethodMayBeStatic
    def end_mlflow_run(self, status: Any = None) -> None:
        pass

    def log_metric(
        self,
        name: str,
        time_diff_in_minutes: float,
        log_level: LogLevel = LogLevel.TRACE,
    ) -> None:
        pass

    def log_param(
        self, key: str, value: str, log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        pass

    def log_params(
        self, params: Dict[str, Any], log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        pass

    # noinspection PyUnusedLocal
    def log_artifact(
        self, key: str, contents: str, folder_path: Optional[str] = None
    ) -> None:
        pass

    def write_to_log(
        self, name: str, message: str = "", log_level: LogLevel = LogLevel.INFO
    ) -> bool:
        return True

    def log_exception(self, event_name: str, event_text: str, ex: Exception) -> None:
        self.log_artifact("_exception.txt", str(ex))
        if self.event_loggers:
            for event_logger in self.event_loggers:
                event_logger.log_exception(
                    event_name=event_name, event_text=event_text, ex=ex
                )

    def log_progress_event(
        self,
        event_name: str,
        current: int,
        total: int,
        event_format_string: str,
        backoff: bool = True,
        log_level: LogLevel = LogLevel.TRACE,
    ) -> None:
        self.logger.debug(event_format_string.format(event_name, current, total))
        if not self.system_log_level or self.system_log_level == LogLevel.INFO:
            if (
                log_level == LogLevel.INFO or log_level == LogLevel.ERROR
            ):  # log only INFO messages
                if self.event_loggers:
                    for event_logger in self.event_loggers:
                        event_logger.log_progress_event(
                            event_name=event_name,
                            current=current,
                            total=total,
                            event_format_string=event_format_string,
                            backoff=backoff,
                        )
        else:  # LOGLEVEL is lower than INFO
            if self.event_loggers:
                for event_logger in self.event_loggers:
                    event_logger.log_progress_event(
                        event_name=event_name,
                        current=current,
                        total=total,
                        event_format_string=event_format_string,
                        backoff=backoff,
                    )

    def log_event(
        self, event_name: str, event_text: str, log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        self.write_to_log(name=event_name, message=event_text)
        if not self.system_log_level or self.system_log_level == LogLevel.INFO:
            if (
                log_level == LogLevel.INFO or log_level == LogLevel.ERROR
            ):  # log only INFO messages
                if self.event_loggers:
                    for event_logger in self.event_loggers:
                        event_logger.log_event(
                            event_name=event_name, event_text=event_text
                        )
        else:
            if self.event_loggers:
                for event_logger in self.event_loggers:
                    event_logger.log_event(event_name=event_name, event_text=event_text)
